/* cap_t.cpp
 *
 * libcap Wrapper for the Microsoft Network Monitor analyzer
 * By Laurent Rabret <laurent.rabret@francetelecom.com>
 * Copyright 2001 FTR&D
 *
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "includes.h"

pcap_t_::pcap_t_(DWORD dNum, DWORD slen, int promisc, int to_ms, char *ebuf) :
			deviceNumber(dNum), snaplen(slen), ReadTimeOut(to_ms),
			MinRange(0xFFFF0000 - to_ms), MaxRange(0xFFFFFFFF - to_ms)
{
		// a MAX_SIZE_OF_BUFFER bytes buffer
		datafield = new unsigned char[MAX_SIZE_OF_BUFFER];
		if (!datafield)
			throw -1;

		if ((hTimeOutEvent = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL)
			throw -2;

		pBuffer = pBufferLast = pRecycler = NULL;
		pIRTC = NULL;
		fieldbegin = fieldend = 0;
		ZeroMemory(szError, sizeof(szError));
		hBlobWorking = NULL;
		lostPacket	= 0;

		if (InitializeLivePCap(promisc, ebuf) == FALSE) {
			// The destructor won't be called so, we must clean up the instance here
			delete datafield;
			throw -1;
		}
		InitializeCriticalSection(&m_CS);
}


pcap_t_::~pcap_t_()
{
	if (pIRTC) {
		// Standard IUnknown call to release the interface
		pIRTC->Release();
		pIRTC = NULL;

	}
	if (pBuffer) {
		delete pBuffer;
		pBuffer = NULL;
	}
	if (pRecycler) {
		delete pRecycler;
		pRecycler = NULL;
	}
	if (datafield) {
		delete datafield;
		datafield = NULL;
	}
	if (hBlobWorking) {
		DestroyBlob(hBlobWorking);
	}
	// Don't delete pBufferLast please !! It's just a shortcut
	DeleteCriticalSection(&m_CS);

	CloseHandle(hTimeOutEvent);
}

BOOL pcap_t_::AddNewFrame(PUCHAR cbuf, unsigned cSize /*off wire*/ , unsigned len /* on wire*/, long time_sec, long time_usec)
{
	lBuffer* pTemp;
	if (pRecycler == NULL) {
		// There 's no unused "frame handler". So, we must create one
		try {
			pTemp = new lBuffer(cbuf, cSize, len, time_sec, time_usec, fieldbegin, fieldend, datafield);
		}
		catch (...) {
			return FALSE;
		}
		if (!pTemp) {
			return FALSE;
		}
	}
	else {
		// There 's at least one unused "frame handler". Let's take it!!
		EnterCriticalSection(&m_CS);
		pTemp= pRecycler;
		pRecycler = pRecycler->next;
		LeaveCriticalSection(&m_CS);

		if (pTemp->InitFields(cbuf, cSize, len, time_sec, time_usec, fieldbegin, fieldend, datafield) == FALSE) {
			// Initialisation failed, we must reconnect pTemp to pRecycler
			EnterCriticalSection(&m_CS);
			pTemp->next = pRecycler;
			pRecycler = pTemp;
			LeaveCriticalSection(&m_CS);
			return FALSE;	
		}
	}
	// Now, we can build the chain
	EnterCriticalSection(&m_CS);
	if (!pBuffer) {
		pBuffer = pTemp;
	}
	else {
		pBufferLast->next = pTemp;
	}
	LeaveCriticalSection(&m_CS);
	pBufferLast = pTemp;
	return TRUE;
}

/*
Description
	The FindAnalyzer subroutine find the network analyzer BLOB (Binary Large Object)
	associated to an id.

Parameters
	number NIC id (this ID is only valuable for this library)

Return Values
	A handle on the BLOB if found, NULL otherwise
*/

HBLOB pcap_t_::FindAnalyzer(DWORD number)
{
	HBLOB hBlobRet = NULL;

	if (number == 0) return NULL;
	
	HBLOB		FilterBlob  = NULL;
	PBLOB_TABLE pBTable		= NULL;

	CreateBlob(&FilterBlob);
	if (FilterBlob == NULL) {
		return NULL;
	}

	if (NMERR_SUCCESS != (SetBoolInBlob( FilterBlob,
			                                OWNER_NPP,
				                            CATEGORY_CONFIG,
					                        TAG_INTERFACE_REALTIME_CAPTURE,
											TRUE))) {
		DestroyBlob(FilterBlob);
		return NULL;
	}

	if (NMERR_SUCCESS == GetNPPBlobTable( FilterBlob, &pBTable)) {
		// number of valid network tables in pBTable->dwNumBlobs
		for (DWORD i=0; i < pBTable->dwNumBlobs; i++) {
			if (i == (number-1)) {
				hBlobRet = pBTable->hBlobs[i];
			}
			else {
				// destroy unused BLOBS
				DestroyBlob(pBTable->hBlobs[i]);
			}
			pBTable->hBlobs[i] = NULL;
		} // end for
		// Destroy the BLOB table
		DestroyBlob(pBTable);
	}
	DestroyBlob(FilterBlob);
	return hBlobRet;
}


BOOL pcap_t_::InitializeLivePCap(int promisc, char *ebuf)
{
	// The first caracter of the "device" param is a number used to quicly find
	// the correct analyzer
	hBlobWorking = FindAnalyzer(deviceNumber);
	if (!hBlobWorking) {
		strcpy(ebuf, "Can't find analyzer anymore");
		return FALSE;
	}

	if ( CreateNPPInterface(hBlobWorking, IID_IRTC, (void**)&pIRTC) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't join the low level network analyzer");
		// We mustn't forget to destroy hBlobWorking because not referred 
		return FALSE;
	}

	// To ask the driver to capture snaplen byte / frame max
	NETWORKINFO nInfo;
	DWORD sLen;
	if (GetNetworkInfoFromBlob(hBlobWorking, &nInfo) == NMERR_SUCCESS) {
		sLen = (snaplen > nInfo.MaxFrameSize) ? sLen = RTC_FRAME_SIZE_FULL : snaplen;
	}
	else {
		sLen = snaplen;
	}

	if (SetDwordInBlob(hBlobWorking, OWNER_NPP, CATEGORY_CONFIG, TAG_FRAME_SIZE, sLen) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't configure snaplen");
		return FALSE;
	}

/*
	// suppressed because not really usefull
	// specify the delay to call the APC
	// must convert from ms to s
	if (SetDwordInBlob(hBlobWorking, OWNER_NPP, CATEGORY_CONFIG, TAG_UPDATE_FREQUENCY, ((ReadTimeOut % 1000) ? 1 : 0) + ReadTimeOut/1000) != NMERR_SUCCESS) {
		strcpy(ebuf, "Can't set callback delay");
		return FALSE;
	}
*/

	// specify whether the card must/mustn't be in promiscuous mode
	if (!promisc) {
		// leave the NIC in non-promiscuous mode
		// by default, NM changes the NIC to promiscuous mode
		// I don't master this part and don't understand why the next call doesn't work...
		// From MS documentation, it is stated the "CAPTUREFILTER_FLAGS_LOCAL_ONLY" const
		// indicates the driver not to work in promiscuous mode. Therefore, the merge of
		// this latter flag + CAPTUREFILTER_FLAGS_INCLUDE_ALL 'should' allow the capture of
		// all packets with a source or dest @ == this NIC. However, it's doesn't work that
		// way...
		// If someone knows the solution thanks to adapt this part.
		// We specify we want all packets and to leave the NIC in non-promiscuous mode
		if (NMERR_SUCCESS != SetNPPEtypeSapFilter(hBlobWorking,
								0, 0, NULL, NULL,
								CAPTUREFILTER_FLAGS_LOCAL_ONLY | CAPTUREFILTER_FLAGS_INCLUDE_ALL,
								NULL)) {
			strcpy(ebuf, "Can't leave the NIC in non-promiscuous mode");
			// NO, we must not retrun FALSE if this call fails
			//return FALSE;
		}

		// Becaue the previous call doesn't work, we specify an address filter
		// with the sourse or dest address == the NIC MAC address
		NETWORKINFO NInfo;
		if (NMERR_SUCCESS == GetNetworkInfoFromBlob(hBlobWorking, &NInfo)) {
			if (!(NInfo.Flags & NETWORKINFO_FLAGS_RAS)) {
				// A dialup connection must not set a filter based on its MAC address
				// We are here because the NIC we are handling is not for DialUp
				// We add a MAC filter...
				ADDRESSTABLE AddTable;
				ZeroMemory(&AddTable, sizeof(AddTable));
				AddTable.nAddressPairs = 2;
				AddTable.AddressPair[0].AddressFlags = ADDRESS_FLAGS_MATCH_DST;
				AddTable.AddressPair[1].AddressFlags = ADDRESS_FLAGS_MATCH_SRC;
	
				memcpy(AddTable.AddressPair[0].DstAddress.MACAddress, 
					NInfo.PermanentAddr, MAC_ADDRESS_SIZE);
				memcpy(AddTable.AddressPair[1].SrcAddress.MACAddress, 
					NInfo.PermanentAddr, MAC_ADDRESS_SIZE);

				if (SetNPPAddressFilterInBlob(hBlobWorking, &AddTable) != NMERR_SUCCESS) {
					strcpy(ebuf, "Can't leave the NIC in non-promiscuous mode");
					return FALSE;
				}
			}
			// else == DialUp Connection have nothing to do
		}
		else {
			strcpy(ebuf, "Can't leave the NIC in non-promiscuous mode");
			return FALSE;
		}
	} // end if (!promisc)
	return TRUE;
}

BOOL pcap_t_::TimeOutElapsed(DWORD TickLimit, DWORD TickCount)
{
	// The count can loop...
	if (TickLimit >= 0xFFFF0000) {
		if ((TickCount >= TickLimit) || (TickCount < MinRange)) {
			return TRUE;
		}
	}
	else {
		if ((TickCount >= TickLimit) && (TickCount < MaxRange)) {
			return TRUE;
		}
	}
	return FALSE;
}

int	pcap_t_::GetFrames(int cnt, pcap_handler callback, u_char *user, BOOL bWithTimeOut)
{
	int i=0;
	DWORD TickLimit = GetTickCount() + ReadTimeOut;
	do {
		// Wait till there's a packet available...
		if (pBuffer)
		{
			lBuffer* temp = pBuffer;
			lBuffer* temp2;

			BOOL bAgain = TRUE;
			do {
				i++;
				callback(user, &temp->pktHead, temp->Buf);
				// Don't need this frame anymore
				fieldbegin = temp->NextBuf;
				EnterCriticalSection(&m_CS);
				temp2 = temp;
				temp = temp->next;

				if (temp == NULL) {
					// there's no frame available
					temp2->next = pRecycler;
					pRecycler	= pBuffer;
					pBuffer		= NULL;
					bAgain		= FALSE;
					LeaveCriticalSection(&m_CS);
					break;
				}

				switch (cnt) {
					case 0 :
						// Get all packets till a timeout elapsed
						if (!bWithTimeOut || !TimeOutElapsed(TickLimit, GetTickCount()))
							break;
					case 1 :
						// case 1 or case 0 with timeout elapsed
						temp2->next = pRecycler;
						pRecycler	= pBuffer;
						pBuffer		= temp;
						bAgain		= FALSE;
						LeaveCriticalSection(&m_CS);
						return i;
					case -1 :
						// infinite loop
						break;
					default :
						--cnt;
				} // switch
				LeaveCriticalSection(&m_CS);
			} while (bAgain);
		} // if (pBuffer)

		DWORD CurrentTime = GetTickCount();
		if (TimeOutElapsed(TickLimit, CurrentTime)) {
			break;
		}

		DWORD dwRetWait = WaitForSingleObjectEx(hTimeOutEvent, (bWithTimeOut == FALSE) ? INFINITE : 
			(CurrentTime <= TickLimit) ?
				(TickLimit - CurrentTime) : 0xFFFFFFFF-CurrentTime+TickLimit, TRUE);
		if ((dwRetWait!= WAIT_OBJECT_0) && (dwRetWait != WAIT_IO_COMPLETION)) {
			// we must exit from the loop
			break;
		}
	} while (1);
	return i;
}

int	pcap_t_::SetFilter(struct bpf_program *fp)
{
	int		ret = 0;
	HRESULT hRes;
	
	if (!pIRTC) {
		strcpy(szError, "Real Time analyser uninitialized");
		return -1;
	}
	// Let's start the analysis
	// we first set the filter...
	if (pIRTC->Connect(hBlobWorking, NULL, FrameProc,
                                (LPVOID)this, NULL) != NMERR_SUCCESS )
	{
		// the connect failed
		strcpy(szError, "NIC connection failed");
        return -1;
    }
	// ...and run
	if ((hRes = pIRTC->Start()) != NMERR_SUCCESS) {
		switch (hRes) {
			case NMERR_CAPTURE_PAUSED	: strcpy(szError, "Can't start. The capture is in pause mode");
				break;
			case NMERR_CAPTURING		: strcpy(szError, "The capture is already started");
				break;	
			case NMERR_NOT_CONNECTED	: strcpy(szError, "The analyzer is not connected to the network");
				break;
			case NMERR_NOT_REALTIME		: strcpy(szError, "Can't use this analyzer in real time");
				break;
			default :
				strcpy(szError, "Unexpected error");
		}
		ret = -1;
	}
	return ret;
}



int	pcap_t_::DataLink()
{
	DWORD ret;
	DWORD dwMacType;
	if (GetNPPMacTypeAsNumber(hBlobWorking, &dwMacType) == NMERR_SUCCESS){
		switch(dwMacType) {
			case MAC_TYPE_UNKNOWN  : ret = 0; break; // not good be to display something
			case MAC_TYPE_ETHERNET : ret = 1; break;
			case MAC_TYPE_TOKENRING: ret = 6; break; 
			case MAC_TYPE_FDDI	   : ret = 10; break;
			default				   : ret = 0;		 // not good also...
		}
	}
	return ret;
}


int	pcap_t_::Stats(struct pcap_stat *ps)
{
	STATISTICS stat;
	if (!pIRTC) {
		strcpy(szError, "Unexpected field in the pcap struct");
		return -1;
	}
	if  (pIRTC->GetTotalStatistics(&stat,TRUE) == NMERR_SUCCESS)
	{
		ps->ps_recv		= stat.TotalFramesSeen;
		ps->ps_drop		= stat.TotalFramesDropped + lostPacket;
		// Not so sure for ps_ifdrop. Is it dropped by filters ?
		ps->ps_ifdrop	= stat.TotalFramesSeen - stat.TotalFramesFiltered; 
	}
	return 0;
}


void pcap_t_::Stop()
{
	if (pIRTC) {
		pIRTC->Stop();
		pIRTC->Disconnect();
	}
}


char *pcap_t_::LookupDev(char *errbuf)
{
	static CHAR driver[4096];
	PCHAR pRet = driver;
	
	HBLOB		FilterBlob  = NULL;
	PBLOB_TABLE pBTable		= NULL;

	try {
		CreateBlob(&FilterBlob);
		if (FilterBlob == NULL) {
			strcpy(errbuf, "Lack of system resources");
			return NULL;
		}

		if (NMERR_SUCCESS != (SetBoolInBlob( FilterBlob,
				                                OWNER_NPP,
					                            CATEGORY_CONFIG,
						                        TAG_INTERFACE_REALTIME_CAPTURE,
												TRUE))) {
			DestroyBlob(FilterBlob);
			strcpy(errbuf, "Can't find suitable network interface");
			return NULL;
		}

		if (NMERR_SUCCESS == GetNPPBlobTable( FilterBlob, &pBTable))
		{
			driver[0] = '\0';
			int base = 0;

			// number of valid network tables in pBTable->dwNumBlobs
			for (DWORD i=0; i < pBTable->dwNumBlobs; i++) {

				PCHAR TempString;
				
				// 1st, we retreive MAC type (Ethernet, Token Ring etc...
				// Note that a dialup PPP connection will have an Ethernet MAC type
				// due to the way WAN NDIS works.
				// We MUST get this parameter
				if (NMERR_SUCCESS == GetStringFromBlob(pBTable->hBlobs[i],
												OWNER_NPP,
												CATEGORY_NETWORKINFO,
												TAG_MACTYPE,
												(const char**)&TempString)) {
					wsprintf(driver+base, "%d_%s", i+1, TempString);
					base += strlen(driver+base);

					
					// Now, let's find if this blob is for RAS connections
					DWORD dwFlag;
					if (NMERR_SUCCESS == GetDwordFromBlob(pBTable->hBlobs[i],
													OWNER_NPP,
													CATEGORY_NETWORKINFO,
													TAG_FLAGS,
													&dwFlag)) {
						if (dwFlag & NETWORKINFO_FLAGS_RAS) {
							strcpy(driver+base, "_DIALUP");
							base += strlen(driver+base);
						}
					}

					// At last, we get the MAC address
					if (NMERR_SUCCESS == GetStringFromBlob(pBTable->hBlobs[i],
													OWNER_NPP,
													CATEGORY_LOCATION,
													TAG_MACADDRESS,
													(const char**)&TempString)) {
						wsprintf(driver+base, "_%s", TempString);
						base += strlen(driver+base);
					}
					
				}
				base += 1+strlen(driver+base); // +1 to indicate next Network Card
				// Don't need this blob anymore
				DestroyBlob(pBTable->hBlobs[i]);
				pBTable->hBlobs[i] = NULL;
			} // end for
			// must complete the string with a double '\0'
			*(driver+base) = '\0';
		}
		else {
			strcpy(errbuf, "Can't find network driver to analyze");
			pRet = NULL;
		}
		DestroyBlob(FilterBlob);
		FilterBlob = NULL;

		// We don't need this BLOB table anymore
		DestroyBlob(pBTable);
		pBTable = NULL;

	}
	catch(...)
	{
		strcpy(errbuf, "Exception caught");
		pRet = NULL;
		if (FilterBlob) {
			DestroyBlob(FilterBlob);
		}
		if (pBTable) {
			DestroyBlob(pBTable);
		}
	}
	return pRet;
}

/*
Description
	The FrameProc callback function is automatically called by the network analyser driver
	when one of the following condition in encountered:
	1) the buffer is full
	2) the timeslot elapsed and the buffer is not null

Parameters
	Event: descriptor on all captured frames

Return Values
	A network monitor error message. Here, we always return a success status except
	when an exception occurred
*/

DWORD WINAPI  pcap_t_::FrameProc( UPDATE_EVENT Event )
{
	DWORD ret = NMERR_SUCCESS;
	__try {
		// User parameter is actually a pointer to the pcap struct
		pcap_t_* pcap = reinterpret_cast<pcap_t_*>(Event.lpUserContext);

		LPFRAMETABLE lpFrameTable = Event.lpFrameTable;
	
		// Walk through frames
		// The frame table can wrap the indices
	    for ( unsigned Count = lpFrameTable->StartIndex; 
					Count != lpFrameTable->EndIndex; 
					   (Count == lpFrameTable->FrameTableLength ) ? Count=0: Count++ )
		{	
		    LPFRAME_DESCRIPTOR lpFrameDesc = &lpFrameTable->Frames[Count];
			
			// lpFrameDesc->TimeStamp specifies the system time, in microseconds, when the
			// frame was captured. It is a "pure" 64 bits value
			// We need to change reference from January, 1st 1601 to January, 1st 1970 = 11644473600 seconds if I'm right!!
			DWORD time_sec  = (ULONG) (lpFrameDesc->TimeStamp / (__int64)1000000 - 11644473600);
			DWORD time_usec = (ULONG) (lpFrameDesc->TimeStamp % (__int64)1000000);
			if (pcap->AddNewFrame(lpFrameDesc->FramePointer, lpFrameDesc->nBytesAvail, lpFrameDesc->FrameLength, time_sec, time_usec) == FALSE) {
				// we must update the lost packets counter
				pcap->lostPacket += (Count > lpFrameTable->EndIndex) ?
					(2 + lpFrameTable->FrameTableLength - Count + lpFrameTable->EndIndex) :
					(1 + lpFrameTable->EndIndex - Count);
				if (Count != lpFrameTable->StartIndex) {
					// There 's at least one packet registered
					pcap->SignalNewPackets();
				}
				__leave;
			}
		} // end for
		// We can signal the app some packets are available
		pcap->SignalNewPackets();
	}
	__except(EXCEPTION_EXECUTE_HANDLER) {
		// This exception might occured only is the IRTC is disconnected
		ret = NMERR_DISCONNECTED;
	}
	return ret;
}
